package org.jastadd.relast.compiler;

import org.jastadd.option.ValueOption;
import org.jastadd.relast.ast.GrammarFile;
import org.jastadd.relast.ast.Program;
import org.jastadd.relast.parser.RelAstParser;
import org.jastadd.relast.scanner.RelAstScanner;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Optional;

public abstract class RelAstProcessor extends AbstractCompiler {

  protected ValueOption optionOutputBaseDir;
  protected ValueOption optionInputBaseDir;

  public RelAstProcessor(String name, boolean jastAddCompliant) {
    super(name, jastAddCompliant);
  }

  protected static boolean isGrammarFile(String fileName) {
    String extension = fileName.subSequence(fileName.lastIndexOf('.'), fileName.length()).toString();
    return extension.equals(".relast") || extension.equals(".ast");
  }

  @Override
  protected void initOptions() {
    optionOutputBaseDir = addOption(new ValueOption("outputBaseDir", "base directory for generated files"));
    optionInputBaseDir = addOption(new ValueOption("inputBaseDir", "base directory for input files"));
    super.initOptions();
  }

  @Override
  protected int compile() throws CompilerException {
    final Path inputBasePath;
    if (optionInputBaseDir.isMatched()) {
      inputBasePath = Paths.get(optionInputBaseDir.value()).toAbsolutePath();
    } else {
      inputBasePath = Paths.get(".").toAbsolutePath();
      printMessage("No input base dir is set. Assuming current directory '" + inputBasePath.toAbsolutePath().toString() + "'.");
    }

    if (!inputBasePath.toFile().exists()) {
      printMessage("Input path '" + inputBasePath.toAbsolutePath().toString() + "' does not exist. Exiting...");
      System.exit(-1);
    } else if (!inputBasePath.toFile().isDirectory()) {
      printMessage("Input path '" + inputBasePath.toAbsolutePath().toString() + "' is not a directory. Exiting...");
      System.exit(-1);
    }

    final Path outputBasePath;
    if (optionOutputBaseDir.isMatched()) {
      outputBasePath = Paths.get(optionOutputBaseDir.value()).toAbsolutePath();
    } else {
      throw new CompilerException("No output base dir is set.");
    }

    if (outputBasePath.toFile().exists() && !outputBasePath.toFile().isDirectory()) {
      printMessage("Output path '" + inputBasePath.toAbsolutePath().toString() + "' exists, but is not a directory. Exiting...");
    }

    printMessage("Running " + getName());

    // gather all files
    Collection<Path> inputFiles = new ArrayList<>();
    getConfiguration().getFiles().forEach(name -> relativizeFileName(inputBasePath, Paths.get(name)).ifPresent(inputFiles::add));


    Program program = parseProgram(inputFiles);

    return processGrammar(program, inputBasePath, outputBasePath);
  }

  protected abstract int processGrammar(Program program, Path inputBasePath, Path outputBasePath) throws CompilerException;

  private Optional<Path> relativizeFileName(Path inputBasePath, Path filePath) {
    if (filePath.isAbsolute()) {
      if (filePath.startsWith(inputBasePath)) {
        return Optional.of(filePath.relativize(inputBasePath));
      } else {
        printMessage("Path '" + filePath + "' is not contained in the base path '" + inputBasePath + "'.");
        return Optional.empty();
      }
    } else {
      return Optional.of(inputBasePath.resolve(filePath));
    }
  }

  protected void printMessage(String message) {
    System.out.println(message);
  }

  protected void writeToFile(Path path, String str) throws CompilerException {
    //noinspection ResultOfMethodCallIgnored
    path.getParent().toFile().mkdirs(); // create directory structure if necessary
    try (PrintWriter writer = new PrintWriter(path.toFile())) {
      writer.print(str);
    } catch (Exception e) {
      throw new CompilerException("Could not write to file " + path, e);
    }
  }

  private Program parseProgram(Collection<Path> inputFiles) {
    Program program = new Program();

    RelAstParser parser = new RelAstParser();

    inputFiles.stream().filter(path -> isGrammarFile(path.toString())).forEach(
        path -> {
          try (BufferedReader reader = Files.newBufferedReader(path)) {
            RelAstScanner scanner = new RelAstScanner(reader);
            GrammarFile inputGrammar = (GrammarFile) parser.parse(scanner);
            inputGrammar.setFileName(path.toString());
            program.addGrammarFile(inputGrammar);
            inputGrammar.treeResolveAll();
          } catch (IOException | beaver.Parser.Exception e) {
            printMessage("Could not parse grammar file " + path);
            e.printStackTrace();
          }
        }
    );

    return program;
  }
}

