package org.jastadd.relast.tests;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import org.assertj.core.util.Files;
import org.jastadd.relast.tests.config.Configuration;
import org.junit.jupiter.api.Assertions;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;

import static org.junit.jupiter.api.Assertions.assertTrue;

public class RelAstProcessorTestBase {

  protected static int runProcess(File workingDirectory, List<String> command, StringBuilder outStringBuider, StringBuilder errStringBuilder) throws IOException, InterruptedException {

    File outFile = Files.newTemporaryFile();
    File errFile = Files.newTemporaryFile();

    ProcessBuilder pb = new ProcessBuilder(command).
        directory(workingDirectory)
        .redirectOutput(outFile)
        .redirectError(errFile);

    Process p = pb.start();
    try {
      p.waitFor();
    } catch (InterruptedException e) {
      if (Thread.interrupted())  // Clears interrupted status!
        throw e;
    }

    try (BufferedReader outReader = new BufferedReader(new FileReader(outFile))) {
      String outLine;
      while ((outLine = outReader.readLine()) != null) {
        outStringBuider.append(outLine).append("\n");
      }
    }

    try (BufferedReader errReader = new BufferedReader(new FileReader(errFile))) {
      String errLine;
      while ((errLine = errReader.readLine()) != null) {
        errStringBuilder.append(errLine).append("\n");
      }
    }

    return p.exitValue();
  }

  protected static int runJavaProcess(Class<?> klass, File workingDirectory, List<String> args, StringBuilder outStringBuider, StringBuilder errStringBuilder) throws IOException, InterruptedException {
    String javaHome = System.getProperty("java.home");
    String javaBin = javaHome + File.separator + "bin" + File.separator + "java";
    String classpath = System.getProperty("java.class.path");
    String className = klass.getName();

    List<String> command = new LinkedList<>();
    command.add(javaBin);
    command.add("-cp");
    command.add(classpath);
    command.add(className);
    if (args != null) {
      command.addAll(args);
    }

    System.out.println("Running java -jar -cp [...] " + className + " " + args.stream().reduce((s1, s2) -> s1 + " " + s2).orElse(""));

    return runProcess(workingDirectory, command, outStringBuider, errStringBuilder);
  }

  protected void directoryTest(Class<?> mainClass, Path dir) throws IOException, InterruptedException {
    dir = dir.toAbsolutePath();
    Path configFile = dir.resolve("config.yaml");
    ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
    List<Configuration> configs = mapper.readValue(configFile.toFile(), new TypeReference<List<Configuration>>() {
    });

    for (Configuration config : configs) {

      StringBuilder outBuilder = new StringBuilder();
      StringBuilder errBuilder = new StringBuilder();
      int returnValue = runJavaProcess(mainClass, dir.toFile(), Arrays.asList(config.getArgs()), outBuilder, errBuilder);
      String out = outBuilder.toString();
      String err = errBuilder.toString();

      System.out.println(out);
      System.err.println(err);

      if (config.shouldFail()) {
        Assertions.assertNotEquals(0, returnValue, "Zero return value of preprocessor for negative test.");
      } else {
        Assertions.assertEquals(0, returnValue, "Non-Zero return value of preprocessor for positive test.");
      }

      for (String errMatchString : config.getErrMatches()) {
        if (!err.matches(errMatchString)) {
          Assertions.fail("Error stream does not match '" + errMatchString + "'");
        }
      }

      for (String errContainsString : config.getErrContains()) {
        if (!err.contains(errContainsString)) {
          Assertions.fail("Error stream does not contain '" + errContainsString + "'");
        }
      }

      for (String outMatchString : config.getOutMatches()) {
        if (!out.matches(outMatchString)) {
          Assertions.fail("Error stream does not match '" + outMatchString + "'");
        }
      }

      for (String outContainsString : config.getOutContains()) {
        if (!out.contains(outContainsString)) {
          Assertions.fail("Error stream does not contain '" + outContainsString + "'");
        }
      }
    }
  }

  protected void ensureOutputDir(String outputDir) {
    File outputDirFile = Paths.get(outputDir).toFile();
    if (outputDirFile.exists()) {
      assertTrue(outputDirFile.isDirectory());
      if (Objects.requireNonNull(outputDirFile.list(), "Could not read output directory").length != 0) {
        System.out.println("output directory is not empty!");
      }
    } else {
      assertTrue(outputDirFile.mkdirs());
    }
  }
}
